/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include "pic-processor.h"


#ifndef __14_BIT_PROCESSORS_H__
#define __14_BIT_PROCESSORS_H__

     // forward references
class _14bit_processor;
class IOPIN;

extern instruction *disasm14 (_14bit_processor *cpu,unsigned int inst);


//#include "14bit-registers.h"
//#include "14bit-instructions.h"
#include "pic-processor.h"

class _14bit_processor : public pic_processor
{

public:

#define EEPROM_SIZE              0x40
#define INTERRUPT_VECTOR         4
#define WDTE                     4


  PCLATH       pclath;

  INTCON       *intcon;

  EECON1       eecon1;
  EECON2       eecon2;
  EEDATA       eedata;
  EEADR        eeadr;

  EEPROM       eeprom;


  virtual void create_symbols(void);
//  virtual void load_hex(char *hex_file);
  void interrupt(void);
  virtual void por(void);
  virtual void create(void);// {return;};
  virtual PROCESSOR_TYPE isa(void){return _14BIT_PROCESSOR_;};
  virtual PROCESSOR_TYPE base_isa(void){return _14BIT_PROCESSOR_;};
  virtual instruction * disasm (unsigned int inst)
    {
      return disasm14(this, inst);
    }

  // Declare a set of functions that will allow the base class to
  // get information about the derived classes. NOTE, the values returned here
  // will cause errors if they are used -- the derived classes must define their
  // parameters appropriately.
  virtual void create_sfr_map(void) { return;};

  // Return the portion of pclath that is used during branching instructions
  virtual unsigned int get_pclath_branching_jump(void)
    {
      return ((pclath.value & 0x18)<<8);
    }

  // Return the portion of pclath that is used during modify PCL instructions
  virtual unsigned int get_pclath_branching_modpcl(void)
    {
      return((pclath.value & 0x1f)<<8);
    }

  virtual void option_new_bits_6_7(unsigned int);

//  _14bit_processor(void);

};

#define cpu14 ( (_14bit_processor *)cpu)

class PORTB : public IOPORT
{
public:
#define intedg_MASK (1<<6)
#define rbpu_MASK   (1<<7)

  unsigned int rbpu;
  unsigned int intedg;

  void rbpu_intedg_update(unsigned int);
  unsigned int get(void);
  void setbit(unsigned int bit_number, bool new_value);
};

class PORTA : public IOPORT
{
public:

  void setbit(unsigned int bit_number, bool new_value);

};


class _14bit_18pins
{
public:

  INTCON       intcon_reg;
  PORTA        porta;
  IOPORT_TRIS  trisa;

  PORTB        portb;
  IOPORT_TRIS  trisb;

  void create_iopin_map(IOPIN_map ** iopin_map_ptr, int * num_of_iopins_ptr);

};

#endif
