/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

//
// commands.cc
//
//  All of the gpsim command-line commands are defined in here. 
// Also, putting eveything here helps keep the interface to
// input.cc cleaner. (input.cc is a hack of the readline example
// c-program 'fileman.c'.)
//


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>

#define whitespace(c) (((c) == ' ') || ((c) == '\t'))

#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <string.h>
#include <unistd.h>

#include <iostream.h>
#include <iomanip.h>
#include <g++/string>
#include <g++/list>
#include <g++/vector>

#include "14bit-processors.h"
#include "breakpoints.h"
#include "stimuli.h"
#include "symbol.h"
#include "commands.h"
#include "gpsim_def.h"

pic_processor * Command::cpu = NULL;

//int address = 0;
instruction *inst;
//unsigned int value;
unsigned int current_base=0;
//unsigned int last_breakpoint;
unsigned int quit_gpsim = 0;

int argc;
char *argv[MAX_ARGUMENTS];


list <Command *> command_list;
list <Command *> :: iterator cl_iterator;
 
extern void process_command_file(char  *f);

struct cmd_option {
  char *name;
  int  type;
  int  id_mask;
};

/* Look up NAME as the name of a command, and return a pointer to that
   command.  Return a NULL pointer if NAME isn't a command name. */
Command *
find_command (string name)
{
  register int i;

  //  cout << "searching for " << name << '\n';

  for (cl_iterator = command_list.begin();  cl_iterator != command_list.end(); cl_iterator++)
    {
      Command *t = *cl_iterator;
      if ( t->name == name)
	{
	  //	  cout << "found it\n";
	  return (t);
	}
    }
  return ((Command *)NULL);
}


/* Strip whitespace from the start and end of STRING. */
void stripwhite (char *string)
{
  register int i = 0;

  while (whitespace (string[i]))
    i++;

  if (i)
    strcpy (string, string + i);

  i = strlen (string) - 1;

  while (i > 0 && whitespace (string[i]))
    i--;

  string[++i] = '\0';
}




// The following routine will convert the first number it finds in
// a string to an unsigned long integer. All of the hard work is done
// in the library function strtoul (string to unsigned long).

unsigned long get_number_in_string(char *number_string)
{
  unsigned long retval = 0;

  cout << "converting " <<  number_string;
  errno = 0;

  if( (*number_string >= '0') && (*number_string <=9))
    retval = strtoul(number_string, NULL, current_base);
  else
    errno = 1;

  cout << " got " << retval << " and error " << errno << '\n';
  return(retval);
}


//
// c_base
//
// this is not really a command...
// it's used to manipulate the static information in the command class
//

class c_base : public Command
{
public:

  new_cpu(pic_processor *p) 
  { 
    cpu = p;
    trace.switch_cpus(cpu);
  };

} com_base;

class c_attach : public Command
{
public:

  c_attach(void)
  {
    command_state = NO_COMMAND;

    name = string("attach");

    brief_doc = string("Attach stimuli to nodes");

    long_doc = string ("attach node1 stimulus1 [stimulus2 stimulu_N]\
\n\t  attach is used to define the connections between stimuli and nodes.\
\n\tAt least one node and one stimulus must be specified. If more stimuli\
\n\tare specified then they will all be attached to the node.\
\n\n\texamples:
\n\n\tnode pin2pin_test                  // Define a new node.\
\n\tattach pin2pin_test porta4 portb0  // Connect two I/O pins to the node.\
\n\tnode                               // Display the new \"net list\".\
\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    if(argc < 2)
      {
	cout << "-- at least two arguments are needed [help attach]\n";
	return command_state;
      }

    string s = string(argv[1]);
    Stimulus_Node *sn = find_node (s);

    if(sn)
      {
	int i=2;

	stimulus *st;
	while(i <= argc)
	  {
	    s = string(argv[i++]);
	    st = find_stimulus(s);
	    if(st)
	      sn->attach_stimulus(st);
	  }
      }

    return command_state;
  };

} com_attach;


class c_break : public Command
{
public:

  int last_breakpoint;

  c_break(void)
  {
    command_state = NO_COMMAND;

    name = string("break");

    brief_doc = string("Set a break point");

    long_doc = string ("break [c e | w | r | wv | rv | wdt [location] [value] ]\n\n\
\toptions:\n\
\t\tc   - cycle\n\
\t\te   - execution\n\
\t\tw   - write\n\
\t\tr   - read\n\
\t\twv  - write value\n\
\t\trv  - read value\n\
\t\twdt - wdt timeout\n\
\t\t    - no argument, display the break points that are set.
\texamples:\n\
\t\tbreak e 0x20     // set an execution break point at address 0x20\n\
\t\tbreak wv 0x30 0  // break if a zero is written to register 0x30\n\
\t\tbreak c 1000000  // break on the one million'th cycle\n\
\t\tbreak            // display all of the break points\n\
\n");
  };
  
  //
  // Set a break point. This will set an execution break point at the
  // address specified in *arg. If arg is empty, then the most recent
  // address will be reused. I'm not sure if that's desirable...
  //

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    unsigned int a,b,c;

    command_state = NO_COMMAND;
    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }


    // Convert the arguments (if there are any) into options and numbers

    if(argc >= 2)
      {
	char *arg = argv[1];                    // first argument

	// 2nd argument should be either a number or an address label.
	symbol *s = symbol_table.find(argv[2]);
	if(s!= NULL)
	  {
	    errno = 0;
	    if(s->isa() == SYMBOL_CONSTANT)
	      {
		constant_symbol *cs = (constant_symbol *)s;
		a = cs->val;
	      } 
	    else if (s->isa() == SYMBOL_ADDRESS)
	      {
		address_symbol *as = (address_symbol *)s;
		a = as->val;
	      } 
	    else if (s->isa() == SYMBOL_REGISTER)
	      {
		register_symbol *rs = (register_symbol *)s;
		a = rs->reg->address;
	      } 
	  }
        else 
	  {
	    errno = 0;
	    a = strtoul(argv[2], 0, current_base);  // 2nd argument should be a number
	  }

	int e1 = errno;

	if(argc >= 3)
	  b = strtoul(argv[3], 0, current_base);  // 3rd argument should be a number too

	if(errno || e1)
	  {
	    cout << "I was expecting numbers for those arguments.\nNo break was set.\n";
	    return command_state;
	  }

	command_state = NO_COMMAND;   // Assume that we succeed in setting a break point

	if(*arg >= 'A')
	  {
	    switch (*arg)
	      {
	      case 'c':    // Cycle break point
		last_breakpoint = bp.set_cycle_break(cpu,a);

		if(last_breakpoint < MAX_BREAKPOINTS)
		  cout << "break at cycle: " << a << " break #: " <<  last_breakpoint << '\n';
		else
		  cout << "failed to set cycle break\n";

		break;

	      case 'e':    // Execution break point

		c = bp.set_execution_break(cpu, a);
		if(c < MAX_BREAKPOINTS)
		  {
		    last_breakpoint = c;
		    cout << "break at address: " << a << " break #: " <<
		      last_breakpoint << '\n';
		  }
		else
		  cout << "failed to set execution break (check the address)\n";
		break;

	      case 'r':     // Break on register read
		if(*(arg+1) == '\0')
		  {
		    b = bp.set_read_break(cpu, a);
		    if(b < MAX_BREAKPOINTS)
		      cout << "break when register " << a << " is read.\n" << 
			"bp#: " << b << '\n';
		  }
		else if(*(arg+1) == 'v')
		  {
		    c = bp.set_read_value_break(cpu, a,b);
		    if(c < MAX_BREAKPOINTS)
		      cout << "break when " << b << " is read from register " << a << '\n' << 
			"bp#: " << c << '\n';
		  }
		break;

	      case 'w':
		if(*(arg+1) == '\0')
		  {
		    b = bp.set_write_break(cpu, a);
		    if(b < MAX_BREAKPOINTS)
		      cout << "break when register " << a << " is written. bp#: " << b << '\n';
		  }
		else if(*(arg+1) == 'v')
		  {
		    c = bp.set_write_value_break(cpu, a,b);
		    if(c < MAX_BREAKPOINTS)
		      cout << "break when " << b << " is written to register " << a << 
			". bp#: " << c << '\n';
		  }

		break;

	      default:
		command_state = NO_COMMAND;
		cout << "valid options are e r w rv wv\n";
	      }
	  }
      }
    else if (argc == 1)
      {
	if(strcmp(argv[1],"wdt") == 0)
	  {
	    b = bp.set_wdt_break(cpu);
	    if(b < MAX_BREAKPOINTS)
	      cout << "break when wdt times out.  " <<
		"bp#: " << b << '\n';
	  }

      }
    else
      bp.dump();  // No arguments, so just dump all of the break points

    return command_state;

  };


} com_break;




class c_help : public Command
{
public:

  c_help(void)
  {
    command_state = NO_COMMAND;

    name = string("help");

    brief_doc = string("Type help \"command\" for more help on a command");

    long_doc = string ("\ngpsim\tis a software simulator for the Microchip PIC microcontrollers\n\
\tPlease refer to the distributed README files for more information\n\
\tType help \"command\" for more help on a command\n"
);
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    Command *t;

    command_state = NO_COMMAND;

    // If there is no argument, then print the short command summary
    if(!argc)
      {
	for (cl_iterator = command_list.begin();  cl_iterator != command_list.end(); cl_iterator++)
	  {
	    t = *cl_iterator;
	    
	    cout << t->name;
	    int l = 15 - t->name.length();
	    for(int i=0; i<l; i++)
	      cout << ' ';

	    cout << t->brief_doc << '\n';
	  }
      }

    // Otherwise, see if the argument is a valid command
    else 
      {
	for(int i=1; i<=argc; i++)
	  {
	    if( (t=find_command(argv[i])) != NULL ) 
	      {
		cout << t->long_doc << '\n';
	      }
	    else
	      {
		cout << argv[i] << " is not a valid gpsim command. Try these instead:\n";

		int commands_printed_in_this_line = 0;
		for (cl_iterator = command_list.begin();  cl_iterator != command_list.end(); cl_iterator++)
		  {
		    t = *cl_iterator;
	    
		    cout << t->name << "\t";
		    if(++commands_printed_in_this_line > 6 )
		      cout << '\n';
		  }
		cout << '\n';

	      }
	  }
      }
    return command_state;
  };

} com_help;



class c_break_clear : public Command
{
public:

  c_break_clear(void)
  {
    command_state = NO_COMMAND;

    name = string("clear");

    brief_doc = string("Remove a break point");

    long_doc = string ("clear bp_number\n\
where bp_number is the number assigned to the break point\n\
when it was created. (type \"break\" without any arguments to\n\
display the currently set break points.\n\
");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    unsigned long a,b;

    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }

    // Convert the arguments (if there are any) into options and numbers

    if(argc >= 1)
      {
	errno = 0;
	a = strtoul(argv[1], 0, current_base);  // Get the bp number
	if(!errno)
	  bp.clear(a);
      }

    return command_state;
  };

} com_break_clear;

class c_disassemble : public Command
{
public:

  c_disassemble(void)
  {
    command_state = NO_COMMAND;

    name = string("disassemble");

    brief_doc = string("Disassemble the current cpu");

    long_doc = string ("\ndisassemble [[start | length] [end]]\n\n\
\t no arguments:  disassembles 10 instructions before and 5 after the pc.\n\
\t one argument:  disassemble [length] instructions after the pc.\n\
\ttwo arguments:  disassemble from [start] to [end].\n\
\n\
");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    unsigned long start,end;

    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }

    // Convert the arguments (if there are any) into options and numbers

    if(argc >= 1)
      {
	errno = 0;
	start = strtoul(argv[1], 0, current_base);
	if(errno)
	  start = 10;
      }
    else
      start = 10;
      
    if(argc >= 2)
      {
	errno = 0;
	end = strtoul(argv[2], 0, current_base);
	if(errno)
	  end = 5;
      }
    else
      end = 5;

    if(argc == 0)
      cpu->disassemble(cpu->pc.value - start, cpu->pc.value + end);
    else if(argc == 1)
      cpu->disassemble(cpu->pc.value, cpu->pc.value + start);
    else
      cpu->disassemble(start, end);

    return command_state;
  };

} com_disassemble;


class c_dump : public Command
{
public:

  c_dump(void)
  {
    command_state = NO_COMMAND;

    name = string("dump");

    brief_doc = string("Display either the RAM or EEPROM");

    long_doc = string ("dump [r | e]\n
\tdump r or dump with no options will display all of the file\n\
\tregisters and special function registers.\n\
\tdump e will display the contents of the eeprom (if the pic
\tbeing simulated contains any\n\
\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {

    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }

    bool dump_ram = 0;
    if(argc==1)
      {
	if(*argv[1] == 'e')
	  cpu14->eeprom.dump();
	else
	  dump_ram = 1;
      }
    
    if(argc != 1 || dump_ram)
      cpu->dump_registers();

    return command_state;


  };

} com_dump;

class c_echo : public Command
{
public:

  c_echo(void)
  {
    command_state = NO_COMMAND;

    name = string("echo");

    brief_doc = string("echo \"text\"");

    long_doc = string ("echo \"text\" - useful for command files\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    int i=1;

    while(i <= argc)
      {
	
	cout << argv[i++] << ' ';
      }
    cout << '\n';

    return command_state;
  };

} com_echo;

class c_list : public Command
{
public:
  int starting_line, ending_line, file_id,pc,total_lines;
  bool have_pc,have_sl,have_el,have_id;

  c_list(void)
  {
    file_id = 0;
    total_lines = 10;

    command_state = NO_COMMAND;

    name = string("list");

    brief_doc = string("Display source and list files");

    long_doc = string("list [[s | l] [*pc] [line_number1 [,line_number2]]]\
\n\n\tDisplay the contents of source and list files.\
\n\tWithout any options, list will use the last specified options.\
\n\tlist s will display lines in the source (or .asm) file.\
\n\tlist l will display lines in the .lst file\
\n\tlist *pc will display either .asm or .lst lines around the\
\n\t   value specified by pc (e.g. list *20 will list lines around\
\n\t   address 20)\
\n\tline_number1, line_number2 - specify the list range.\
\n\n\tExamples:\
\n\tlist s *0x3a -5 5\
\n\t  will list 11 lines (5 before, 5 after, & 1 at) around addr 3a\
\n\tlist\
\n\t  repeat the last list except use the current pc as the reference.\
\n\tlist l\
\n\t  will list lines from .lst file around the current pc.\
\n
");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    unsigned int n;

    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }

    // Convert the arguments (if there are any) into options and numbers

    if(argc == 0)
      {
	pc = cpu->pc.value;
	starting_line =  -5;
	ending_line =  +5;
      }
    else
      {
	n = 1;
	have_pc = 0;
	// There's at least one argument
	if(*argv[n] >= 'l')
	  {
	    file_id = 99; // a bogus value

	    if(strlen(argv[n]) == 1)
	      {
		if(*argv[n] == 'l')
		  file_id = 1;
		else if(*argv[n] == 's')
		  file_id = 0;
	      }
	    if(file_id == 99)
	      {
		file_id = 0;
		cout << "? valid options are s or l or a line number\n";
		return command_state;
	      }
	    else
	      n++;
	  }

	if(n <= argc)
	  {
	    if(*argv[n] =='*')
	      {
		char *s = argv[n++];
		s++;
		pc = strtoul(s, 0, current_base);
		have_pc = 1;
		n++;
	      }
	  }
	if(!have_pc)
	  pc = cpu->pc.value;

	if(n <= argc)
	  {
	    errno = 0;
	    starting_line = strtoul(argv[n++], 0, current_base);
	    if(errno)
	      starting_line =  -5;
	    ending_line = starting_line + total_lines;

	  }

	if(n <= argc)
	  {
	    errno = 0;
	    ending_line = strtoul(argv[n++], 0, current_base);
	    if(errno)
	      ending_line =  5;
	    if(ending_line < starting_line)
	      ending_line = starting_line;
	    total_lines = ending_line - starting_line;
	  }

      }

    cpu->list(file_id,pc,starting_line,ending_line);

    return command_state;
  };

} com_list;

class c_load : public Command
{
public:

  // Keep track of the number of times the load command is called.
  // The MAX value is arbitrarily large enough to handle normal cases,
  // but is small enough to handle infinite recursion.

  int recursive_depth,i;
  const int MAX_RECURSIVE_DEPTH = 100; 

  c_load(void)
  {
    recursive_depth = 0;

    command_state = NO_COMMAND;

    name = string("load");

    brief_doc = string("Load either a hex,command, or .cod file");

    long_doc = string ("load  h | c | s  file_name\
\n\n\tload either a hex file, command file, or a symbol file.\
\n\t(Byte Craft's .cod files are the only symbol files that\
\n\tare recognized.)\
\n\n\tExample:\
\n\t  load s perfect_program.cod\
\n\t    will load the symbol file perfect_program.cod\
\n\t    note that the .cod file contains the hex stuff\
\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {

    if(++recursive_depth > MAX_RECURSIVE_DEPTH)
      {
	cout << "Error: 'load' was called way too many times...\n";
	recursive_depth--;
	return command_state;
      }

    if(argc == 2)
      {
	char *arg = argv[1];

	if( *(arg + 1) == 0)   // single character option
	  {
	    //	    FILE *f = NULL;

	    if(*arg == 'h')
	      {
		cout << "hex file\n";
		if(NULL == cpu)
		    cout << "(load) no cpu has been selected\n";
		else
		  cpu->load_hex(argv[2]);

	      }
	    else if(*arg == 'c')
	      {
		cout << "command file\n";
		//f = fopen(argv[2],"r");
		process_command_file(argv[2]);
	      }
	    else if(*arg == 's')
	      {
		cout << "symbol file\n";
		if(i=load_symbol_file(&cpu, argv[2]))
		  {
		    cout << "found a fatal error in the symbol file " << argv[2] << '\n';
		    display_symbol_file_error(i);
		  }
		else
		  {
		    com_base.new_cpu(cpu);
		  }
	      }
	  }
      }

    recursive_depth--;
    return command_state;
  };

} com_load;



class c_node : public Command
{
public:

  c_node(void)
  {
    command_state = NO_COMMAND;

    name = string("node");

    brief_doc = string("Add or display stimulus nodes");

    long_doc = string ("node [new_node1 new_node2 ...]\
\n\t If no new_node is specified then all of the nodes that have been\
\n\tdefined are displayed. If a new_node is specified then it will be\
\n\tadded to the node list. See the \"attach\" and \"stimulus\" commands\
\n\tto see how stimuli are added to the nodes.\
\n\n\texamples:
\n\n\tnode              // display the node list\
\n\tnode n1 n2 n3     // create and add 3 new nodes to the list\
\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {

    if(argc == 0)
      {
	dump_node_list();
	return command_state;
      }
    int i=1;

    while(i <= argc)
      {
	//create a new node and add it to the node list.
	Stimulus_Node *sn = new Stimulus_Node(argv[i++]);
      }

    return command_state;
  };

} com_node;


class c_processor : public Command
{
public:

  c_processor(void)
  {
    command_state = NO_COMMAND;

    name = string("processor");

    brief_doc = string("Select & Display processors");

    long_doc = string ("processor [new_processor_type new_processor_name] | [list]\
\n\tIf no new processor is specified, then the currently defined processor(s)\
\n\twill be displayed. To see a list of the processors supported by gpsim,\
\n\ttype 'processor list'.  To define a new processor, specify the processor\
\n\ttype and name.\
\n\n\texamples:
\n\n\tprocessor               // Display the processors you've already defined.\
\n\tprocessor list          // Display the list of processors supported.\
\n\tprocessor p16cr84 fred  // Create a new processor.\
\n\tprocessor p16c74 wilma  // and another.\
\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    int i;

    if(argc == 0)
      {
	dump_processor_list();
      }

    else if(argc == 1)
      {
	if(strcmp(argv[1],"list") == 0)
	  display_available_processors();
	else
	  {
	    // if the one arguement is a processor, go ahead and add it
	    // to the processor list. The name will be the same as the
	    // processor type.
	    i = find_in_available_processor_list(argv[1]);
	    if (i>=0)
	      com_base.new_cpu(add_processor(argv[1], argv[1]));
	    else
	      cout << argv[1] << " makes no sense to me.\nTry help processor.\n";
	  }

      }

    else if(argc == 2)
      {
	//create a new processor and add it to the processor list.
	com_base.new_cpu(add_processor(argv[1], argv[2]));
      }
    else
      {
	cout << "too many options specified. Try 'help processor'\n";
      }

    return command_state;
  };

} com_processor;

class c_quit : public Command
{
public:

  c_quit(void)
  {
    command_state = NO_COMMAND;

    name = string("quit");

    brief_doc = string("Quit gpsim");

    long_doc = string ("Quit gpsim\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    quit_gpsim = 1;
    return command_state;
  };

} com_quit;

class c_reset : public Command
{
public:

  c_reset(void)
  {
    command_state = NO_COMMAND;

    name = string("reset");

    brief_doc = string("Reset all or parts of the simulation");

    long_doc = string ("Reset\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {

    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }

    cpu->reset(POR_RESET);

    return command_state;
  };

} com_reset;

class c_run : public Command
{
public:

  c_run(void)
  {
    command_state = NO_COMMAND;

    name = string("run");

    brief_doc = string("Execute the pic program");

    long_doc = string ("run\n\
\tStart simulating and don't stop until a break is encountered.\n\
\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }

    cout << "running...\n";


    cpu->run();

    return command_state;

  };

} com_run;


class c_step : public Command
{
public:

  c_step(void)
  {
    command_state = NO_COMMAND;

    name = string("step");

    brief_doc = string("Execute one or more instructions.");

    long_doc = string ("\nstep [over | n]\n\n\
\t    no arguments:  step one instruction.\n\
\tnumeric argument:  step a number of instructions\n\
\t \"over\" argument:  step over the next instruction\n\
\n\
");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    unsigned int n;

    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }

    // Convert the arguments (if there are any) into options and numbers

    n = 1;

    if(argc == 1)
      {
	if(strcmp(argv[1],"over") == 0)
	  {
	    cpu->step_over();
	    return command_state;
	  }

	errno = 0;
	n = strtoul(argv[1], 0, current_base);  // Get the number of steps
	if(errno)
	  n = 1;
      }

    cpu->step(n);

    return command_state;

  };

} com_step;


// This command needs a massive overhaul. Expect it to be moved really soon!

#define  STIM_NOTHING          0
#define  STIM_PERIOD          (1 << 0)
#define  STIM_PHASE           (1 << 1)
#define  STIM_HIGH_TIME       (1 << 2)
#define  STIM_INITIAL_STATE   (1 << 3)
#define  STIM_START_CYCLE     (1 << 4)
#define  STIM_DATA            (1 << 5)
#define  STIM_IOPORT          (1 << 6)
#define  STIM_ASY             (1 << 7)
#define  STIM_SQW             (1 << 8)
#define  STIM_NAME            (1 << 9)
#define  STIM_TRI             (1 << 10)
#define  STIM_AD              (1 << 11)
#define  STIM_ANALOG          (1 << 12)
#define  STIM_DIGITAL         (1 << 13)


static struct cmd_option stimulus_options[] = {
  { "period"       , 0, STIM_PERIOD         },
  { "initial_state", 0, STIM_INITIAL_STATE  },
  { "phase"        , 0, STIM_PHASE          },
  { "high_time"    , 0, STIM_HIGH_TIME      },
  { "start_cycle"  , 0, STIM_START_CYCLE    },
  { "name"         , 0, STIM_NAME           },
  { "analog"       , 0, STIM_ANALOG         },
  { "digital"      , 0, STIM_DIGITAL        }
};

static int num_stimulus_options = sizeof (stimulus_options) / sizeof(cmd_option);

class c_stimulus : public Command
{
public:
  unsigned int options_entered, valid_options;

  const unsigned int
  SQW_OPTIONS = STIM_SQW | STIM_PERIOD | STIM_PHASE | STIM_HIGH_TIME | STIM_START_CYCLE;
  const unsigned int
  ASY_OPTIONS = STIM_ASY | STIM_PERIOD | STIM_PHASE | STIM_HIGH_TIME | STIM_START_CYCLE | STIM_DATA;
  const unsigned int
  TRI_OPTIONS = STIM_TRI | STIM_PERIOD | STIM_PHASE | STIM_HIGH_TIME | STIM_START_CYCLE;


  asynchronous_stimulus *asy;
  square_wave *sqw;
  ioport_symbol *is;
  unsigned int period,
    phase,
    high_time,
    start_cycle,
    bit_pos,
    states,
    initial_state,
    time_flag,
    digital;

  char n[30];

  vector<int>  temp_array;


  c_stimulus(void)
  {
    command_state = NO_COMMAND;
    valid_options = STIM_NOTHING;

    name = string("stimulus");

    brief_doc = string("Create a stimulus");

    long_doc = string ("\nstimulus [[type] options]\n\
\tstimulus will create a signal that can be tied to an io port.
\tNote that in most cases it is easier to create a stimulus
\tfile then to type all this junk by hand.\n\n\
\t  Supported stimuli:\n\n\
\tsquare_wave | sqw  [period p] [high_time h] [phase ph] [initial_state i]\n\
\t  port port_name bit_pos end\n\
\t\t  creates a square wave with a period of \"p\" cpu cycles.\n\
\t\t  If the high time is specified then that's the number of cycles\n\
\t\t  the square wave will be high.\n\
\t\t  The phase is with respect to the cpu's cycle counter.\n\
\t\t  The \"port_name\" and \"bit_pos\" describe where the stimulus\n\
\t\t  will be attached.
\tasynchronous_stimulus | asy  [period p] [phase ph]  [initial_state i]
\t  d0 [d1 d2 ... dn] port port_name bit_pos end\n\
\t\t  creates an asynchronous square wave with a period of \"p\" cpu
\t\t  cycles.  The phase is with respect to the cpu's cycle counter.\n\
\t\t  The \"port_name\" and \"bit_pos\" describe where the stimulus\n\
\t\t  will be attached.

\n\texamples:\n\n\
\t  stimulus sqw period 200 high_time 20 phase 60 port portb 0 end 
\t  create a square wave stimulus that repeats every 200 cpu cycles, 
\t  is high for 20 cpu cycles (and low for 200-20=180 cycles). The 
\t  first rising edge will occur at cycle\n\
\t  60, the second at 260, . . . Bit 0 of portb will receive the stimulus.\n\
\n");
};

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }

    int i=0,t;

    if(command_state == NO_COMMAND)
      {
	// initialize some default values.
	period = 1000;
	phase = 0;
	high_time = 500;
	initial_state = 1;
	start_cycle = 0;
	bit_pos = 0;
	states = 0;
	time_flag = 1;
	digital = 1;

	if(temp_array.size())
	  temp_array.erase(temp_array.begin(), temp_array.end());

	*n = 0;

	if(argc>0)
	  {
	    // check for a valid stimulus type
	    if((strcmp(argv[1],"sqw") == 0) || (strcmp(argv[1],"square_wave") == 0))
	      {
		command_state = COMMAND_PENDING;
		valid_options = SQW_OPTIONS;
		options_entered = STIM_SQW;
		i = 2;
	      }
	    else if((strcmp(argv[1],"asy") == 0) || (strcmp(argv[1],"asynchronous_stimulus") == 0))
	      {
		command_state = COMMAND_PENDING;
		valid_options = ASY_OPTIONS;
		options_entered = STIM_ASY;
		i = 2;
	      }
	    else if((strcmp(argv[1],"tri") == 0) || (strcmp(argv[1],"triangle_wave") == 0))
	      {
		command_state = COMMAND_PENDING;
		valid_options = TRI_OPTIONS;
		options_entered = STIM_TRI;
		i = 2;
	      }
	    else if( (strcmp(argv[1],"d") == 0) || (strcmp(argv[1],"dump") == 0) )
	      {  // dump the stimulus list
		dump_stimulus_list();
		command_state = NO_COMMAND;
		return command_state;
	      }
	    else
	      {
		cout << argv[1] << " is not a valid stimulus type\n";
		return command_state;
	      }
	  }
	else
	  {
	    dump_stimulus_list();
	    command_state = NO_COMMAND;
	    return command_state;
	  }
      }

    if(command_state == COMMAND_PENDING)
      {
	while(i<=argc)
	  {
	    int j = 0;

	    do {

	      if(strcmp(argv[i], stimulus_options[j].name) == 0) {
		options_entered |= stimulus_options[j].id_mask;

		switch (stimulus_options[j].id_mask) {
		case STIM_PERIOD:
		  period = strtoul(argv[++i], 0, current_base);
		  cout << "found period in case statement";
		case STIM_INITIAL_STATE:
		  double d;

		  i++;
		  
		  if( (argv[i][1] == 'x') || (argv[i][1] == 'X') )
		    d = double(strtoul(argv[i], 0, current_base));
		  else
		    d = strtod(argv[i], 0);
		  
		  initial_state = int(double(MAX_ANALOG_DRIVE)*255.0*d);
		  break;

		case STIM_PHASE:
		  phase = strtoul(argv[++i], 0, current_base);
		  break;
		case STIM_HIGH_TIME:
		  high_time = strtoul(argv[++i], 0, current_base);
		  break;
		case STIM_START_CYCLE:
		  start_cycle = strtoul(argv[++i], 0, current_base);
		  break;
		case STIM_NAME:
		  strcpy(n,argv[++i]);
		  break;
		case STIM_ANALOG:
		  digital = 0;
		  break;
		case STIM_DIGITAL:
		  digital = 1;
		  break;
		default:
		  cout << stimulus_options[j].name << " is not in this case statement\n";
		}
		//i++;
		break; // out of the 'j' loop
	      }
	    } while(++j < num_stimulus_options);


	    if(j<num_stimulus_options)
	      cout << "option must have been found\n";

	    else if(strcmp(argv[i],"port") == 0)
	      {
		symbol *s = symbol_table.find(argv[++i]);
		if(s)
		  {

		    if(s->isa() == SYMBOL_IOPORT)
		      {
			is = (ioport_symbol *)(s);

			options_entered |= STIM_IOPORT;

			++i;
			if(isdigit(*argv[i]))
			  bit_pos = strtoul(argv[i], 0, current_base);

		      }
		    else
		      cout << argv[i-1] << " is not an ioport.\n";
		  }
		else
		  cout << "symbol " << argv[i] << " was not found.\n";

	      }
	    else if(strcmp(argv[i],"end") == 0)
	      {

		if(options_entered & STIM_SQW)  // && (options_entered & STIM_IOPORT))
		  {
		    if(options_entered & STIM_NAME)
		      sqw = new square_wave(period,high_time,phase,n);
		    else
		      sqw = new square_wave(period,high_time,phase);
		    if (options_entered & STIM_IOPORT)
			is->ioport->attach_stimulus(sqw, bit_pos);

		  }
		else if(options_entered & STIM_TRI)
		  {
		    triangle_wave *tri;

		    if(options_entered & STIM_NAME)
		      tri = new triangle_wave(period,high_time,phase,n);
		    else
		      tri = new triangle_wave(period,high_time,phase);
		    if (options_entered & STIM_IOPORT)
			is->ioport->attach_stimulus(tri, bit_pos);

		  }
		else if(options_entered & STIM_ASY) // && (options_entered & STIM_IOPORT))
		  {
		    if(temp_array.size())
		      {

			if(options_entered & STIM_NAME)
			  asy =  new asynchronous_stimulus(n);
			else
			  asy =  new asynchronous_stimulus();

			asy->digital = digital;
			asy->max_states =  temp_array.size()/2;
			asy->transition_cycles = new int[ asy->max_states];
			asy->values            = new int[ asy->max_states];

			for(int j=0; j<=2*asy->max_states; j++)
			  {
			    if(j&1)
			      asy->values[j>>1] = temp_array[j];
			    else
			      asy->transition_cycles[j>>1] = temp_array[j];
			  }

			asy->cpu = cpu;
			asy->initial_state = initial_state;
			asy->period = period;
			asy->start_cycle = start_cycle;
			if (options_entered & STIM_IOPORT)
			    is->ioport->attach_stimulus(asy, bit_pos);

			asy->start();
		      }
		  }

		command_state = NO_COMMAND;

	      }
	    //else if(isdigit(*argv[i]) && (options_entered & STIM_ASY))
	    //  {
	    //    temp_array.push_back( strtoul(argv[i], 0, current_base) );
	    //  }
	    else if( (options_entered & STIM_ASY) && 
		     (isdigit(*argv[i]) || 
		      (*argv[i]=='-') ||
		      (*argv[i]=='+') ||
		      (*argv[i]=='.')) )
	      {
		errno = 0;
		double d;

		// Check to see if a hex value was entered:

		if( (argv[i][1] == 'x') || (argv[i][1] == 'X') )
		  d = double(strtoul(argv[i], 0, current_base));
		else
		  d = strtod(argv[i], 0);

		if(errno == 0)
		  {
		    // The numeric value was successfully converted to a double.
		    // If it was a 'time' value (i.e. the time when the asynchronous
		    // stimulus will change states) then place the value ( d ) into
		    // the temp array without any scaling. If d represents the value
		    // the asynchronous stim will assume, then place it into the temp
		    // array with scaling.

		    int t;
		    if(time_flag&1)
		      t = int(d);
		    else
		      t = int(double(MAX_ANALOG_DRIVE)*255.0*d);

		    temp_array.push_back(t);
		    // cout << "double " << d << " int " << t << " time_flag " << time_flag << '\n';

		    // Toggle the flag so that the next through we convert the other

		    time_flag ^= 1;

		  }
		else
		  {
		    cout << "Couldn't convert analog value: " << argv[i] << '\n';
		  }
	      }
	    else
	      {
		cout << argv[i] << " is not a valid stimulus option.\n";
		command_state = NO_COMMAND;
		return command_state;
	      }
	    i++;
	  }


      }
    else
      cout << "Hmmm, command_state is screwed.\n";

    return command_state;

  };

} com_stimulus;


class c_symbol : public Command
{
public:

  c_symbol(void)
  {
    command_state = NO_COMMAND;

    name = string("symbol");

    brief_doc = string("Add or display symbols");

    long_doc = string ("symbol [symbol_name [symbol_type]]\n\
\n\tIf no options are supplied, the entire symbol table will be\
\n\tdisplayed. If only the symbol_name is provided, then only\
\n\tthat symbol will be displayed. If both the symbol_name and\
\n\tthe symbol_type are supplied, then a new symbol will be\
\n\tadded to the symbol table.\
\n\n\tValid symbol types:\
\n\t  ioport | iop\
\n");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    if(NULL == cpu)
      {
	cout << "no cpu has been selected\n";
	return command_state;
      }

    switch(argc)
      {
      case 0:
	symbol_table.dump_all();
	break;
      case 1:
	symbol_table.dump_one(argv[1]);
	break;

      case 2:
	symbol_table.add(argv[1], argv[2]);
	break;

      default:
	cout << "Too many arguements\n";
      }

    return command_state;

  };

} com_symbol;


class c_trace : public Command
{
public:

  c_trace(void)
  {
    command_state = NO_COMMAND;

    name = string("trace");

    brief_doc = string("Dump the trace history");

    long_doc = string ("\ntrace [dump_amount]\n\
\ttrace will print out the most recent \"dump_amount\" traces.\n\
\tIf no dump_amount is specified, then the entire trace buffer\n\
\twill be displayed.
");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {

    errno = 0;
    int n = 0;

    command_state = NO_COMMAND;

    if(argc == 1)
      n = strtoul(argv[1], 0, current_base);  // Get the dump amount

    if( (argc != 1) || (errno != 0))
      trace.dump();
    else
      trace.dump(n);

    return command_state;

  };

} com_trace;

class c_version : public Command
{
public:

  c_version(void)
  {
    command_state = NO_COMMAND;

    name = string("version");

    brief_doc = string("Display the gpsim's version");

    long_doc = string("Display the gpsim's version");
  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {
    cout << GPSIM_VERSION << '\n';

    return command_state;
  };

} com_version;


class c_x : public Command
{
public:

  c_x(void)
  {
    command_state = NO_COMMAND;

    name = string("x");

    brief_doc = string("examine and/or modify memory");

    long_doc = string ("\nx [file_register] [new_value]\n\
\toptions:\n\
\t\tfile_register - ram location to be examined or modified.\n\
\t\tnew_value - the new value written to the file_register.\n\
\t\tif no options are specified, then the entire contents\n\
\t\tof the file registers will be displayed (dump).\n\
");

  };

  virtual COMMAND_MODES execute(int argc, char *argv[])
  {

    errno = 0;
    int reg,ov,nv;

    if(argc >= 1)
      {
	if(isdigit(*argv[1]))
	  {
	    reg = strtoul(argv[1], 0, current_base);  // Get the file_register
	  }
	else
	  {
	    errno =1;
	    symbol *s = symbol_table.find(argv[1]);
	    if(s!= NULL) {
	      errno = 0;
	      if(s->isa() == SYMBOL_CONSTANT)
		{
		  constant_symbol *cs = (constant_symbol *)s;
		  reg = cs->val;
		} else if (s->isa() == SYMBOL_REGISTER) {
		  s->print();
		  register_symbol *rs = (register_symbol *)s;
		  reg = rs->reg->address;
		} else if (s->isa() == SYMBOL_IOPORT) {
		  ioport_symbol *iop = (ioport_symbol *)s;
		  reg = iop->ioport->address;
		}
	    }
	  }

	if( (errno != 0) || (reg>=256))
	  {
	    cout << "bad file register\n";
	    return command_state;
	  }

	ov =  cpu->registers[reg]->get_value();
	nv = -1;

	if(argc>=2)
	  {
	    nv = strtoul(argv[2], 0, current_base);  // Get the new_value
	    if( (errno != 0) || (nv>=256))
	      {
		cout << "bad new_value\n";
		return command_state;
	      }

	    // write the new value to the file register:
	    cpu->registers[reg]->put(nv);

	    // read it back (some of the file registers have bit's that can't be changed)
	    nv = cpu->registers[reg]->get_value();
	  }

	cout<< cpu->registers[reg]->name() << '(' << hex << reg << ')';

	if( (ov == nv) || (argc ==1))
	  cout << " is " << ov << '\n';
	else
	  cout << " was " << ov << " now is " << nv << '\n';

      }
    else
      cpu->dump_registers();

    return command_state;
  };

} com_x;

com_newcpu(char *arg)
{

  printf("switching cpus\n");

}


void initialize_commands(pic_processor *c)
{
  com_base.new_cpu(c);

  //  address = 0;
  //value = 0;
  current_base = 0;
  //  last_breakpoint = 0;

  command_list.push_back(&com_attach);
  command_list.push_back(&com_break);
  command_list.push_back(&com_break_clear);
  command_list.push_back(&com_disassemble);
  command_list.push_back(&com_dump);
  command_list.push_back(&com_echo);
  command_list.push_back(&com_help);
  command_list.push_back(&com_list);
  command_list.push_back(&com_load);
  command_list.push_back(&com_node);
  command_list.push_back(&com_processor);
  command_list.push_back(&com_quit);
  command_list.push_back(&com_reset);
  command_list.push_back(&com_run);
  command_list.push_back(&com_step);
  command_list.push_back(&com_stimulus);
  command_list.push_back(&com_symbol);
  command_list.push_back(&com_trace);
  command_list.push_back(&com_version);
  command_list.push_back(&com_x);

  cl_iterator = command_list.begin();

}
