/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


//
// p16x6x
//
//  This file supports:
//    P16C61


#include <stdio.h>
#include <iostream.h>
#include <g++/string>

#include "symbol.h"

#include "p16x6x.h"


void P16C61::create_sfr_map(void)
{
  cout << "creating c61 registers\n";

  add_file_registers(0x0c, 0x2f, 0x80);

 
  add_sfr_register(&indf,  0x80);
  add_sfr_register(&indf,  0x00);

  add_sfr_register(&tmr0,  0x01);
  add_sfr_register(&option_reg,  0x81, 0xff);

  add_sfr_register(&pcl,    0x02, 0);
  add_sfr_register(&status, 0x03, 0x18);
  add_sfr_register(&fsr,    0x04);
  alias_file_registers(0x02,0x04,0x80);

  add_sfr_register(&porta,  0x05);
  add_sfr_register(&trisa,  0x85, 0x3f);

  add_sfr_register(&portb,  0x06);
  add_sfr_register(&trisb,  0x86, 0xff);

  add_sfr_register(&pclath, 0x8a, 0);
  add_sfr_register(&pclath, 0x0a, 0);

  add_sfr_register(&intcon_reg, 0x8b, 0);
  add_sfr_register(&intcon_reg, 0x0b, 0);

  sfr_map = NULL;
  num_of_sfrs = 0;

  intcon = &intcon_reg;

}


void P16C61::create_symbols(void)
{

  symbol_table.add_ioport(portb.cpu, &portb);
  symbol_table.add_ioport(porta.cpu, &porta);


}

//
// External initialization of the general purpose register map
//

//const GPR_map P16C61::gpr_map[];


P16C61::P16C61(void)
{
  cout << "c61 constructor, type = " << isa() << '\n';

  create_iopin_map(&iopin_map, &num_of_iopins);
  create_iopins(iopin_map, num_of_iopins);

  _14bit_processor::create();

  create_sfr_map();

  name_str = "16c61";
}


//
//------------------------------------------------------------------------
//
//

void _14bit_40pins::create_iopin_map(IOPIN_map ** iopin_map_ptr, int * num_of_iopins_ptr)
{

  IOPIN_map im[] =
  {
    // pin, port, bit in port, type
    {2, &porta, 0, BI_DIRECTIONAL},       //0
    {3, &porta, 1, BI_DIRECTIONAL},
    {4, &porta, 2, BI_DIRECTIONAL},
    {5, &porta, 3, BI_DIRECTIONAL},
    {6, &porta, 4, OPEN_COLLECTOR},

    {33, &portb, 0, BI_DIRECTIONAL_PU},   //5
    {34, &portb, 1, BI_DIRECTIONAL_PU},
    {35, &portb, 2, BI_DIRECTIONAL_PU},
    {36, &portb, 3, BI_DIRECTIONAL_PU},
    {37, &portb, 4, BI_DIRECTIONAL_PU},
    {38, &portb, 5, BI_DIRECTIONAL_PU},
    {39, &portb, 6, BI_DIRECTIONAL_PU},
    {40, &portb, 7, BI_DIRECTIONAL_PU},

    {15, &portc, 0, BI_DIRECTIONAL},      //13
    {16, &portc, 1, BI_DIRECTIONAL},
    {17, &portc, 2, BI_DIRECTIONAL},
    {18, &portc, 3, BI_DIRECTIONAL},
    {23, &portc, 4, BI_DIRECTIONAL},
    {24, &portc, 5, BI_DIRECTIONAL},
    {25, &portc, 6, BI_DIRECTIONAL},
    {26, &portc, 7, BI_DIRECTIONAL}

    //{19, &portd, 0, BI_DIRECTIONAL},
    //{20, &portd, 1, BI_DIRECTIONAL},
    //{21, &portd, 2, BI_DIRECTIONAL},
    //{22, &portd, 3, BI_DIRECTIONAL},
    //{27, &portd, 4, BI_DIRECTIONAL},
    //{28, &portd, 5, BI_DIRECTIONAL},
    //{29, &portd, 6, BI_DIRECTIONAL},
    //{30, &portd, 7, BI_DIRECTIONAL},

    //{8, &porte, 0, BI_DIRECTIONAL},
    //{9, &porte, 1, BI_DIRECTIONAL},
    //{10, &porte, 2, BI_DIRECTIONAL},

  };

  // Create a place for permanently storing the I/O pin mapping.

  IOPIN_map *iopin_map = (IOPIN_map *) new char[sizeof (im)];

  int num_of_iopins = sizeof (im) / sizeof (IOPIN_map);

  *iopin_map_ptr = iopin_map;
  *num_of_iopins_ptr = num_of_iopins;

  // Save the map.
  for(int i=0; i<num_of_iopins; i++)
    iopin_map[i] = im[i];


       // ---- Complete the initialization for the I/O Ports

  // Build the links between the I/O Ports and their tris registers.
  porta.tris = &trisa;
  trisa.port = &porta;

  portb.tris = &trisb;
  trisb.port = &portb;

  portc.tris = &trisc;
  trisc.port = &portc;

  // And give them a more meaningful name.
  porta.new_name("porta");
  trisa.new_name("trisa");

  portb.new_name("portb");
  trisb.new_name("trisb");

  portc.new_name("portc");
  trisc.new_name("trisc");

  // Define the valid I/O pins.
  porta.valid_iopins = 0x1f;
  portb.valid_iopins = 0xff;
  portc.valid_iopins = 0xff;

}

void P16C64::create_sfr_map(void)
{
  cout << "creating c64 registers\n";

  add_file_registers(0x20, 0x7f, 0);
  add_file_registers(0xa0, 0xbf, 0);

 
  add_sfr_register(&indf,  0x80);
  add_sfr_register(&indf,  0x00);

  add_sfr_register(&tmr0,  0x01);
  add_sfr_register(&option_reg,  0x81, 0xff);

  add_sfr_register(&pcl,    0x02, 0);
  add_sfr_register(&status, 0x03, 0x18);
  add_sfr_register(&fsr,    0x04);
  alias_file_registers(0x02,0x04,0x80);

  add_sfr_register(&porta,  0x05);
  add_sfr_register(&trisa,  0x85, 0x3f);

  add_sfr_register(&portb,  0x06);
  add_sfr_register(&trisb,  0x86, 0xff);

  add_sfr_register(&portc,  0x07);
  add_sfr_register(&trisc,  0x87, 0x1f);

  //add_sfr_register(&portd,  0x08);
  //add_sfr_register(&trisd,  0x88, 0xff);

  //add_sfr_register(&porte,  0x09);
  //add_sfr_register(&trise,  0x89, 0x07);

  add_sfr_register(&pclath, 0x8a, 0);
  add_sfr_register(&pclath, 0x0a, 0);

  add_sfr_register(&intcon_reg, 0x8b, 0);
  add_sfr_register(&intcon_reg, 0x0b, 0);

  add_sfr_register(&pir1,   0x0c, 0);
  add_sfr_register(&pie1,   0x8c, 0);

  add_sfr_register(&tmr1l,  0x0e, 0);
  add_sfr_register(&tmr1h,  0x0f, 0);

  //add_sfr_register(&pcon,   0x8e, 0);

  add_sfr_register(&t1con,  0x10, 0);
  add_sfr_register(&tmr2,   0x11, 0);
  add_sfr_register(&t2con,  0x12, 0);
  add_sfr_register(&pr2,    0x92, 0xff);

  //add_sfr_register(&sspbuf,  0x13, 0);
  //add_sfr_register(&sspcon,  0x14, 0);

  //add_sfr_register(&sspadd,  0x93, 0);
  //add_sfr_register(&sspstat, 0x94, 0);

  add_sfr_register(&ccpr1l,  0x15, 0);
  add_sfr_register(&ccpr1h,  0x16, 0);
  add_sfr_register(&ccp1con, 0x17, 0);

  sfr_map = NULL;
  num_of_sfrs = 0;

  intcon = &intcon_reg;
  intcon_reg.pir1 = &pir1;
  intcon_reg.pir2 = NULL;

  // Maybe there's a better place for this, but let's go ahead and link all
  // of the registers together (there's probably a better way too) :

  tmr1l.tmr1h = &tmr1h;
  tmr1l.t1con = &t1con;
  tmr1l.pir1  = &pir1;
  tmr1l.ccpcon = &ccp1con;

  tmr1h.tmr1l = &tmr1l;

  t1con.tmr1l = &tmr1l;

  t2con.tmr2  = &tmr2;
  tmr2.pir1   = &pir1;
  tmr2.pr2    = &pr2;
  tmr2.t2con  = &t2con;
  tmr2.ccp1con = &ccp1con;
  tmr2.ccp2con = &ccp2con;
  pr2.tmr2    = &tmr2;


  ccp1con.ccprl = &ccpr1l;
  ccp1con.pir   = &pir1;
  ccp1con.tmr2  = &tmr2;
  ccpr1l.ccprh  = &ccpr1h;
  ccpr1l.tmr1l  = &tmr1l;
  ccpr1h.ccprl  = &ccpr1l;

  portc.ccp1con = &ccp1con;

  ccpr1l.new_name("ccpr1l");
  ccpr1h.new_name("ccpr1h");
  ccp1con.new_name("ccp1con");

  pir1.intcon = &intcon_reg;
  pir1.pie    = &pie1;
  pie1.pir    = &pir1;
  pie1.new_name("pie1");

}

void P16C64::create_symbols(void)
{

  cout << "creating c64 symbols\n";

  symbol_table.add_ioport(porta.cpu, &porta);
  symbol_table.add_ioport(portb.cpu, &portb);
  symbol_table.add_ioport(portc.cpu, &portc);

}



P16C64::P16C64(void)
{
  cout << "c64 constructor, type = " << isa() << '\n';

  create_iopin_map(&iopin_map, &num_of_iopins);
  create_iopins(iopin_map, num_of_iopins);

  sfr_map = NULL;
  num_of_sfrs = 0;
 
  _14bit_processor::create();

  create_sfr_map();

  // Build the links between the I/O Pins and the internal peripherals
  ccp1con.iopin = portc.pins[2];

  name_str = "16c64";

  trace.program_counter (pc.value);

}

//------------------------------------------------------------------------
//
//

void P16C65::create_sfr_map(void)
{

  cout << "creating c65 registers\n";

  add_file_registers(0xc0, 0xff, 0);

  add_sfr_register(&pir2,   0x0d, 0);
  add_sfr_register(&pie2,   0x8d, 0);

  add_sfr_register(&ccpr2l, 0x1b, 0);
  add_sfr_register(&ccpr2h, 0x1c, 0);
  add_sfr_register(&ccp2con, 0x1d, 0);

  intcon_reg.pir2 = &pir2;

  ccp2con.ccprl = &ccpr2l;
  ccp2con.pir   = &pir2;
  ccp2con.tmr2  = &tmr2;
  ccpr2l.ccprh  = &ccpr2h;
  ccpr2l.tmr1l  = &tmr1l;
  ccpr2h.ccprl  = &ccpr2l;

  ccpr2l.new_name("ccpr2l");
  ccpr2h.new_name("ccpr2h");
  ccp2con.new_name("ccp2con");

  pir2.intcon = &intcon_reg;
  pir2.pie    = &pie2;
  pie2.pir    = &pir2;
  pie2.new_name("pie2");

}

void P16C65::create_symbols(void)
{

  cout << "creating c65 symbols\n";

  // There's nothing to create...

}


//------------------------------------------------------------------------
//
// P16C65 constructor
// 
// Note: Since the 'C65 is derived from the 'C64. So before this constructor
// is called, the C64 constructor will be called. Most of the initialization
// is done within the 'C64 constructor.

P16C65::P16C65(void)
{

  cout << "c65 constructor, type = " << isa() << '\n';

  create_sfr_map();

  // Build the links between the I/O Pins and the internal peripherals
  // ccp1con.iopin = portc.pins[2];
  ccp2con.iopin = portc.pins[1];


  name_str = "16c65";

}

// PORTC definition ----- this will be moved later.....

unsigned int PORTC::get(void)
{
  unsigned int old_value;

  old_value = value;

  //cout << "PORTC::get()\n";
  IOPORT::get();

  int diff = old_value ^ value; // The difference between old and new

  // 
  if( diff & CCP1)
    ccp1con->new_edge(value & CCP1);
 
}

void PORTC::setbit(unsigned int bit_number, bool new_value)
{

  unsigned int old_value = value;

  //  cout << "PORTC::setbit() bit " << bit_number << " to " << new_value << '\n';

  IOPORT::setbit( bit_number,  new_value);

  int diff = old_value ^ value; // The difference between old and new

  if( diff & CCP1)
    ccp1con->new_edge(value & CCP1);


}
