/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


//
// p16x8x
//
//  This file supports:
//    PIC16C84
//    PIC16CR84
//    PIC16F84
//    PIC16F83
//    PIC16CR83
//

#include <stdio.h>
#include <iostream.h>
#include <g++/string>

#include "symbol.h"

#include "p16x8x.h"



void _14bit_18pins::create_iopin_map(IOPIN_map ** iopin_map_ptr, int * num_of_iopins_ptr)
{

  IOPIN_map im[] =
  {
    // pin, port, bit in port, type
    {17, &porta, 0, BI_DIRECTIONAL},
    {18, &porta, 1, BI_DIRECTIONAL},
    {1, &porta, 2, BI_DIRECTIONAL},
    {2, &porta, 3, BI_DIRECTIONAL},
    {3, &porta, 4, OPEN_COLLECTOR},

    {6, &portb, 0, BI_DIRECTIONAL_PU},
    {7, &portb, 1, BI_DIRECTIONAL_PU},
    {8, &portb, 2, BI_DIRECTIONAL_PU},
    {9, &portb, 3, BI_DIRECTIONAL_PU},
    {10, &portb, 4, BI_DIRECTIONAL_PU},
    {11, &portb, 5, BI_DIRECTIONAL_PU},
    {12, &portb, 6, BI_DIRECTIONAL_PU},
    {13, &portb, 7, BI_DIRECTIONAL_PU}
  };

  // Create a place for permanently storing the I/O pin mapping.

  IOPIN_map *iopin_map = (IOPIN_map *) new char[sizeof (im)];

  int num_of_iopins = sizeof (im) / sizeof (IOPIN_map);

  *iopin_map_ptr = iopin_map;
  *num_of_iopins_ptr = num_of_iopins;

  // Save the map.
  for(int i=0; i<num_of_iopins; i++)
    iopin_map[i] = im[i];


       // ---- Complete the initialization for the I/O Ports

  // Build the links between the I/O Ports and their tris registers.
  porta.tris = &trisa;
  trisa.port = &porta;

  portb.tris = &trisb;
  trisb.port = &portb;

  // And give them a more meaningful name.
  porta.new_name("porta");
  trisa.new_name("trisa");

  portb.new_name("portb");
  trisb.new_name("trisb");

  // Define the valid I/O pins.
  porta.valid_iopins = 0x1f;
  portb.valid_iopins = 0xff;

}

void P16C8x::create_sfr_map(void)
{

#define UNKNOWN  0

  const SFR_map sm[] =
  {
    {0x00, &indf, UNKNOWN},
    {0x80, &indf, UNKNOWN},

    {0x01, &tmr0, UNKNOWN},
    {0x81, &option_reg, 0xff},

    {0x02, &pcl, 0},
    {0x82, &pcl, 0},

    {0x03, &status, 0x18},
    {0x83, &status, 0x18},

    {0x04, &fsr, UNKNOWN},
    {0x84, &fsr, UNKNOWN},

    {0x05, &porta, UNKNOWN},
    {0x85, &trisa, 0x1f},

    {0x06, &portb, UNKNOWN},
    {0x86, &trisb, 0xff},

    {0x08, &eedata, UNKNOWN},
    {0x88, &eecon1, 0},

    {0x09, &eeadr, UNKNOWN},
    {0x89, &eecon2, UNKNOWN},

    {0x0a, &pclath, 0},
    {0x8a, &pclath, 0},

    {0x0b, &intcon_reg, 0},
    {0x8b, &intcon_reg, 0}
  };


  // Create a place for permanently storing the special function register mapping.

  sfr_map = (SFR_map *) new char[sizeof (sm)];

  num_of_sfrs = sizeof (sm) / sizeof (SFR_map);

  // Save the SFR's.
  for(int i=0; i<num_of_sfrs; i++)
    sfr_map[i] = sm[i];

  intcon = &intcon_reg;
}

void P16C8x::create_symbols(void)
{
  cout << "8x create symbols\n";
  symbol_table.add_ioport(portb.cpu, &portb);
  symbol_table.add_ioport(porta.cpu, &porta);


}

//========================================================================
//
// Pic 16C84 
//


//
// External initialization of the general purpose register map
//

const GPR_map P16C84::gpr_map[];


P16C84::P16C84(void)
{
  cout << "c84 constructor, type = " << isa() << '\n';
  create_sfr_map();
  create_iopin_map(&iopin_map, &num_of_iopins);

  _14bit_processor::create();

  create_iopins(iopin_map, num_of_iopins);

  eeprom.initialize(EEPROM_SIZE,&eecon1, &eecon2, &eedata, &eeadr);
  eeprom.cpu = this;

  name_str = "16c84";
}



//========================================================================
//
// Pic 16F84 
//


//
// External initialization of the general purpose register map
//

const GPR_map P16F84::gpr_map[];


P16F84::P16F84(void)
{
  cout << "f84 constructor, type = " << isa() << '\n';
  create_sfr_map();
  create_iopin_map(&iopin_map, &num_of_iopins);

  _14bit_processor::create();

  create_iopins(iopin_map, num_of_iopins);

  eeprom.initialize(EEPROM_SIZE,&eecon1, &eecon2, &eedata, &eeadr);
  eeprom.cpu = this;

  name_str = "16f84";
}




//========================================================================
//
// Pic 16F83
//


//
// External initialization of the general purpose register map
//

const GPR_map P16F83::gpr_map[];


P16F83::P16F83(void)
{
  cout << "f83 constructor, type = " << isa() << '\n';
  create_sfr_map();
  create_iopin_map(&iopin_map, &num_of_iopins);

  _14bit_processor::create();

  create_iopins(iopin_map, num_of_iopins);

  eeprom.initialize(EEPROM_SIZE,&eecon1, &eecon2, &eedata, &eeadr);
  eeprom.cpu = this;

  name_str = "16f83";
}



