/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

//#include <stdio.h>
#include <stdio.h>
#include <iostream.h>
#include <iomanip.h>
#include <string>
#include <list>

#include "gpsim_def.h"
#include "pic-processor.h"
#include "stimuli.h"
#include "p16x8x.h"
#include "p16x6x.h"
#include "p16x7x.h"
#include "p12x.h"

SIMULATION_MODES simulation_mode;
unsigned int simulation_start_cycle;

extern unsigned int config_word;

//================================================================================
//
// pic_processor
//
// This file contains all (most?) of the code that simulates those features 
// common to all pic microcontrollers.
//
//

//-------------------------------------------------------------------
//
// Define a list for keeping track of the processors being simulated.
// (Recall, gpsim can simultaneously simulate more than one processor.)

list <pic_processor *> processor_list;
list <pic_processor *> :: iterator processor_iterator;


//-------------------------------------------------------------------
//
//
// this sorta sucks, but everytime a new processor is added to the
// gpsim source code, another definition needs to be specified here
// in the 'available_processors' array. Furthermore, a new case 
// statement is also required in 'add_processor'.
//

static const processor_types available_processors[] =
{
  {_PIC_PROCESSOR_, "generic_pic", "generic_pic", "generic_pic", "generic_pic" },
  {_14BIT_PROCESSOR_,"14bit_pic", "14bit_pic", "14bit_pic", "14bit_pic" },
  {_12BIT_PROCESSOR_,"12bit_pic", "12bit_pic", "12bit_pic", "12bit_pic" },
  {_P12C508_,"__12C508", "pic12c508",  "p12c508", "12c508" },
  {_P12C509_,"__12C509", "pic12c509",  "p12c509", "12c509" },
  {_P16C84_, "__16C84",  "pic16c84",   "p16c84",  "16c84"  },
  {_P16CR83_,"__16CR83", "pic16cr83",  "p16cr83", "16cr83" },
  {_P16CR84_,"__16CR84", "pic16cr84",  "p16cr84", "16cr84" },
  {_P16F83_,"__16F83",   "pic16f83",   "p16f83",  "16f83"  },
  {_P16F84_,"__16F84",   "pic16f84",   "p16f84",  "16f84"  },
  {_P16C61_,"__16C61",   "pic16c61",   "p16c61",  "16c61"  },
  {_P16C71_,"__16C71",   "pic16c71",   "p16c71",  "16c71"  },
  {_P16C62_,"__16C62",   "pic16c62",   "p16c62",  "16c62"  },
  {_P16C62A_,"__16C62A", "pic16c62a",  "p16c62a", "16c62a" },
  {_P16CR62_,"__16CR62", "pic16cr62",  "p16cr62", "16cr62" },
  {_P16C63_,"__16C63",   "pic16c63",   "p16c63",  "16c63"  },
  {_P16C64_,"__16C64",   "pic16c64",   "p16c64",  "16c64"  },
  {_P16C64A_,"__16C64A", "pic16c64a",  "p16c64a", "16c64a" },
  {_P16CR64_,"__16CR64", "pic16cr64",  "p16cr64", "16cr64" },
  {_P16C65A_,"__16C65A", "pic16c65a",  "p16c65a", "16c65a" },
  {_P16C65_,"__16C65",   "pic16c65",   "p16c65",  "16c65"  },
  {_P16C74_,"__16C74",   "pic16c74",   "p16c74",  "16c74"  },

};

void dump_processor_list(void)
{

  cout << "Processor List\n";

  bool have_processors = 0;

  for (processor_iterator = processor_list.begin();  processor_iterator != processor_list.end(); processor_iterator++)
    {
      pic_processor *p = *processor_iterator;
      cout << p->name_str << '\n';
      have_processors = 1;
    }

  if(!have_processors)
    cout << "(empty)\n";
}

//-------------------------------------------------------------------
//
//

void display_available_processors(void)
{

  int number_of = sizeof(available_processors) / sizeof(processor_types);
  int i,j,k,l,longest;

  for(i=0,longest=0; i<number_of; i++)
    {
      k = strlen(available_processors[i].names[1]);
      if(k>longest)
	longest = k;
    }

  k=0;
  do
    {

      for(i=0; (i<4) && (k<number_of); i++)
	{
	  cout << available_processors[k].names[1];
	  if(i<3)
	    {
	      l = longest + 2 - strlen(available_processors[k].names[1]);
	      for(j=0; j<l; j++)
		cout << ' ';
	    }
	  k++;
	}
      cout << '\n';
    } while (k < number_of);

}

//-------------------------------------------------------------------
int find_in_available_processor_list(char * processor_type)
{

  int number_of = sizeof(available_processors) / sizeof(processor_types);
  int i,j;

  for(i=0; i<number_of; i++)
    for(j=0; j<4; j++)
      if(strcmp(processor_type,available_processors[i].names[j]) == 0)
	{
	  return(i);
	}

  return(-1);

}
//-------------------------------------------------------------------
pic_processor * add_processor(char * processor_type, char * processor_new_name)
{
  cout << "Trying to add new processor '" << processor_type << "' named '" 
       << processor_new_name << "'\n";

  int i = find_in_available_processor_list(processor_type);

  if(i>= 0)
    {

      pic_processor *p = NULL;

      // sigh, I really, really, really hate this section...
      // All it does is instantiate a new pic processor. We all ready
      // know the name and type of processor we want, however there's no
      // really clean way of indirectly instantiating an object.
      switch(available_processors[i].type)
	{
	case _PIC_PROCESSOR_:
	case _14BIT_PROCESSOR_:
	case _12BIT_PROCESSOR_:
	  cout << "gpsim doesn't support generic processors (should it?).\n";
	  break;

	case _P12C508_:
	  p = new P12C508;
	  break;

	case _P12C509_:
	  p = new P12C509;
	  break;

	case _P16C84_:
	  p = new P16C84;
	  break;

	case _P16CR83_:
	  p = new P16CR83;
	  break;

	case _P16CR84_:
	  p = new P16CR84;
	  break;

	case _P16F83_:
	  p = new P16F83;
	  break;

	case _P16F84_:
	  p = new P16F84;
	  break;

	case _P16C71_:
	  p = new P16C71;
	  break;

	case _P16C61_:
	  p = new P16C61;
	  break;

	case _P16C64_:
	  p = new P16C64;
	  break;

	case _P16C65_:
	  p = new P16C65;
	  break;

	case _P16C74_:
	  p = new P16C74;
	  break;

	default:
	  cout << "error: (BUG) add_processor doesn't have a " << processor_type << " in it's switch block\n";
	  cout << "However, a processor named " << available_processors[i].names[1] << " was found\n";

	}
      if(p)
	{
	  processor_list.push_back(p);
	  cout << processor_type << '\n';
	  return p;
	}
      else
	{
	  cout << " unable to add a processor (BUG?)\n";
	}
    }
  else
    {
      cout << processor_type << " is not a valid processor.\n(try 'processor list' to see a list of valid processors.\n";
    }

  return(NULL);
}


//-------------------------------------------------------------------
//
// sleep - Begin sleeping and stay asleep until something causes a wake
//

void pic_processor::sleep (void)
{
  simulation_mode = SLEEPING;

  do
    {
      cycles.increment();   // burn cycles until something wakes us
    } while(bp.have_sleep()); // %%%FIX ME%%% ===>  && !bp.have_halt());

  simulation_mode = RUNNING;

}

//-------------------------------------------------------------------
//
// run  -- Begin simulating and don't stop until there is a break.
//

void pic_processor::run (void)
{

  simulation_mode = RUNNING;

  // If the first instruction we're simulating is a break point, then ignore it.

  if(program_memory[pc.value]->isa() == instruction::BREAKPOINT_INSTRUCTION)
    {
      simulation_start_cycle = cycles.value;
    }

  do
    {
      do
	{
	  program_memory[pc.value]->execute();
	} while(!bp.global_break);

      if(bp.have_interrupt())
	interrupt();

      if(bp.have_sleep())
	sleep();

    } while(!bp.global_break);

  bp.clear_global();
  trace.dump_last_instruction();

  simulation_mode = STOPPED;

}

//-------------------------------------------------------------------
//
// step - Simulate one (or more) instructions. If a breakpoint is set
// at the current PC, 'step' will go right through it. (That's supposed
// to be a feature.)
//

void pic_processor::step (unsigned int steps)
{


  simulation_mode = SINGLE_STEPPING;
  do
    {

      if(bp.have_sleep())
	{
	  // If we are sleeping, then step one cycle - but don't execute
	  // any code  

	  cycles.increment();
	  trace.dump(1);

	}
      else if(bp.have_interrupt())
	{
	  interrupt();
	}
      else
	{

	  program_memory[pc.value]->execute();
	  trace.dump_last_instruction();

	} 

    }
  while(!bp.have_halt() && --steps>0);

  bp.clear_halt();
  simulation_mode = STOPPED;

}

//-------------------------------------------------------------------
//
// step_over - In most cases, step_over will simulate just one instruction.
// However, if the next instruction is a branching one (e.g. goto, call, 
// return, etc.) then a break point will be set after it and gpsim will
// begin 'running'. This is useful for stepping over time-consuming calls.
//

void pic_processor::step_over (void)
{

  unsigned int saved_pc = pc.value;

  step(1); // Try one step

  if( ! ( (pc.value >= saved_pc) && (pc.value <= saved_pc+2) ) )
    {
      // Set a break point at the instruction just past the one over which we are stepping
      unsigned int bp_num = bp.set_execution_break(this, saved_pc + 1);
      run();
      bp.clear(bp_num);
    }
}

//-------------------------------------------------------------------
//
// reset - reset the pic based on the desired reset type.
//

void pic_processor::reset (RESET_TYPE r)
{
  cout << " --- Reset\n";

  for(int i=0; i<register_memory_size(); i++)
    {
      if(registers[i]->isa() == file_register::SFR_REGISTER)
	{
	  sfr_register *reg = (sfr_register *)registers[i];
	  if(r == POR_RESET)
	    reg->value = reg->por_value;
	  else if (r==WDT_RESET)
	    reg->value = reg->wdtr_value;
	}
    }

  trace.reset(r);
  pc.reset();
  stack.reset();
  bp.clear_global();

  if(r == POR_RESET)
    {
      status.put_TO(1);
      status.put_PD(1);
      cout << "POR\n";
      wdt.initialize( (config_word & WDTE) != 0, nominal_wdt_timeout);
    }
  else if (r==WDT_RESET)
    status.put_TO(0);


}

//-------------------------------------------------------------------
//
// por - power on reset %%% FIX ME %%% needs lots of work...
//

void pic_processor::por(void)
{

  wdt.initialize( (config_word & WDTE) != 0, nominal_wdt_timeout);

}

//-------------------------------------------------------------------
//
// disassemble - Disassemble the contents of program memory from
// 'start_address' to 'end_address'. The instruction at the current
// PC is marked with an arrow '==>'. If an instruction has a break
// point set on it then it will be marked with a 'B'. The instruction
// mnemonics come from the class declarations for each instruction.
// However, it is possible to modify this on a per instruction basis.
// In other words, each instruction in the program memory has it's
// own instantiation. So a MOVWF at address 0x20 is different than
// one at address 0x21. It is possible to change the mnemonic of
// one without affecting the other. As of version 0.0.7 though, this
// is not implemented.
//

void pic_processor::disassemble (int start_address, int end_address)
{
  instruction *inst;
  int use_src_to_disasm =0;

  if(start_address < 0) start_address = 0;
  end_address &= pc.memory_size_mask;

  char str[50];

  for(int i = start_address; i<=end_address; i++)
    {
      if (pc.value == i)
	cout << "==>";
      else
	cout << "   ";

      inst = program_memory[i];

      // Breakpoints replace the program memory with an instruction that has
      // an opcode larger than 16 bits.

      if(program_memory[i]->opcode < 0x10000)
	{
	  cout << ' ';
	}
      else
	{
	  cout << 'B';
	  Breakpoint_Instruction *bpi =  (Breakpoint_Instruction *)program_memory[i];
	  inst = bpi->replaced;
	}

      if((files != NULL) && (use_src_to_disasm))
	{
	  int fid = program_memory[i]->file_id;
	  char buf[256];
	  //cout << " in " << files[fid].name << " @line " 
	  //     << program_memory[i]->src_line << '\n';
	  fseek(files[fid].file_ptr, 
		files[fid].line_seek[program_memory[i]->src_line - 1],
		SEEK_SET);
	  fgets(buf, 256, files[fid].file_ptr);
	  cout << buf;
	}
      else
	cout << hex << setw(4) << setfill('0') << i << "  "
	     << hex << setw(4) << setfill('0') << inst->opcode << "    "
	     << inst->name(str) << '\n';

    }
}

//-------------------------------------------------------------------
//
// pic_processor -- list
//
// Display the contents of either a source or list file
//
void pic_processor::list(int file_id, int pc_val, int start_line, int end_line)
{


  if(number_of_source_files == 0)
    return;

  if(pc_val > program_memory_size())
    return;

  if(program_memory[pc_val]->isa() == instruction::INVALID_INSTRUCTION)
    {
      cout << "There's no code at address 0x" << hex << pc_val << '\n';
      return;
    }

  int line,pc_line;
  if(file_id)
    {
      file_id = lst_file_id;
      line = program_memory[pc_val]->get_lst_line();
      pc_line = program_memory[pc.value]->get_lst_line();
    }
  else
    {
      file_id = program_memory[pc_val]->file_id;
      line = program_memory[pc_val]->get_src_line();
      pc_line = program_memory[pc.value]->get_src_line();
    }

  start_line += line;
  end_line += line;

  if(start_line < 0) start_line = 0;
  if(end_line > files[file_id].max_line)
    end_line = files[file_id].max_line;

  cout << " listing " << files[file_id].name << " Starting line " << start_line
       << " Ending line " << end_line << '\n';

  char buf[256];
  for(int i=start_line; i<=end_line; i++)
    {
      if(!files[file_id].file_ptr)
	files[file_id].file_ptr = fopen(files[file_id].name,"r");
      fseek(files[file_id].file_ptr, 
	    files[file_id].line_seek[i],
	    SEEK_SET);
      fgets(buf, 256, files[file_id].file_ptr);

      if (pc_line == i)
	cout << "==>";
      else
	cout << "   ";

      cout << buf;
    }
}

//-------------------------------------------------------------------
//
// pic_processor -- constructor
//

pic_processor::pic_processor(void)
{

  cout << "pic_processor constructor\n";
  files = NULL;

}


invalid_file_register::invalid_file_register(unsigned int at_address)
{

  char name_str[100];
  sprintf (name_str, "invalid fr  0x%02x", at_address);
  new_name(name_str);
}

//-------------------------------------------------------------------
//
// 
//    create
//
//  The purpose of this member function is to 'create' a pic processor.
// Since this is a base class member function, only those things that
// are common to all pics are created.

void pic_processor :: create (void)
{

  cout << "pic_processor create, type = " << isa() << '\n';

  init_program_memory (program_memory_size());

  init_register_memory (register_memory_size());

  if(num_of_sfrs)
    {
      create_file_registers(get_GPR_map(), GPR_map_size());
      create_sfr_registers ( sfr_map, num_of_sfrs);
    }

  create_invalid_registers ();
  create_symbols();

  // Now, initialize the core stuff:
  pc.cpu = this;
  //  cycles.cpu = this;
  wdt.cpu = this;

  W.new_name("W");
  fsr.new_name("fsr");

  register_bank = &registers[0];  // Define the active register bank 
  W.value = 0;

  set_frequency(10e6);            // 
  nominal_wdt_timeout = 18e-3;    // 18ms according to the data sheet (no prescale)
  Vdd = 5.0;                      // Assume 5.0 volt power supply

  trace.program_counter (pc.value);

  if(num_of_sfrs)
    tmr0.start();

}


//-------------------------------------------------------------------
//
// 
//    create_file_registers
//
//  The purpose of this member function is to allocate memory for the special function
// registers (e.g. status) and the general purpose registers.
//


void pic_processor::create_file_registers (const GPR_map gpr_map[], unsigned int num_of_gprs)
{
  int i, j;

  // Initialize the General Purpose Registers:

  char str[100];
  for (i = 0; i < num_of_gprs; i++)
    {
      for (j = gpr_map[i].start_address; j <= gpr_map[i].end_address; j++)
	{
	  registers[j] = new file_register;
	  if (gpr_map[i].alias_offset)
	    registers[j + gpr_map[i].alias_offset] = registers[j];
	  registers[j]->address = j;
	  registers[j]->break_point = 0;
	  registers[j]->value = 0;
	  registers[j]->symbol_alias = NULL;

	  //The default register name is simply the register's address
	  sprintf (str, "0x%02x", j);
	  registers[j]->new_name(str);

	}
    }


}

//-------------------------------------------------------------------
//
// 
//    add_file_registers
//
//  The purpose of this member function is to allocate memory for the
// general purpose registers.
//

void pic_processor::add_file_registers(unsigned int start_address, unsigned int end_address, unsigned int alias_offset)
{

  int j;

  // Initialize the General Purpose Registers:

  char str[100];
  for (j = start_address; j <= end_address; j++)
    {
      registers[j] = new file_register;
      if (alias_offset)
	registers[j + alias_offset] = registers[j];
      registers[j]->address = j;
      registers[j]->break_point = 0;
      registers[j]->value = 0;
      registers[j]->symbol_alias = NULL;

      //The default register name is simply the register's address
      sprintf (str, "0x%02x", j);
      registers[j]->new_name(str);

    }

}

//-------------------------------------------------------------------
//
// 
//    alias_file_registers
//
//  The purpose of this member function is to alias the
// general purpose registers.
//

void pic_processor::alias_file_registers(unsigned int start_address, unsigned int end_address, unsigned int alias_offset)
{

  int j;

  for (j = start_address; j <= end_address; j++)
    {
      if (alias_offset)
	registers[j + alias_offset] = registers[j];

    }

}

//-------------------------------------------------------------------
//
//
// create_invalid_registers
//
//   The purpose of this function is to complete the initialization
// of the file register memory by placing an instance of an 'invalid
// file register' at each 'invalid' memory location. Most of PIC's
// do not use the entire address space available, so this routine
// fills the voids.
//

void pic_processor::create_invalid_registers (void)
{
  int i;

  // Now, initialize any undefined register as an 'invalid register'
  // Note, each invalid register is given its own object. This enables
  // the simulation code to efficiently capture any invalid register
  // access. Furthermore, it's possible to set break points on
  // individual invalid file registers. By default, gpsim halts whenever
  // there is an invalid file register access.

  for (i = 0; i < FILE_REGISTERS; i++)
    {
      if (NULL == registers[i])
	{
	  registers[i] = new invalid_file_register(i);
	  registers[i]->address = BAD_REGISTER;
	  registers[i]->value = 0;	// unimplemented registers are read as 0
	  registers[i]->symbol_alias = NULL;

	}
    }

}


//-------------------------------------------------------------------
//
// add_sfr_register
//
// The purpose of this routine is to add one special function register
// to the file registers. If the sfr has a physical address (like the 
// status or tmr0 registers) then a pointer to that register will be
// placed in the file register map. 

void pic_processor::add_sfr_register(sfr_register *reg, unsigned int addr,unsigned int por_value=0)
{

  if(addr < register_memory_size())
    {
      registers[addr] = reg;
      registers[addr]->address = addr;
    }

  reg->value       = por_value;
  reg->por_value   = por_value;
  reg->break_point = 0;              // %%% FIX ME %%% Is this still needed?
  reg->cpu         = this;

}

//-------------------------------------------------------------------
//
//
// create_sfr_registers      %%% FIX ME %%% Warning this routine will be obsolete soon!
//
// The purpose of this routine is to create the special function register
// portion of the file register memory map.
//

void pic_processor::create_sfr_registers (const SFR_map sfr_map[], unsigned int num_of_sfrs)
{

  if(num_of_sfrs == 0)
    return;

  int i;

  // Initialize the Special Function Registers:
  //cout << "create_sfrs   num_of_sfrs " << num_of_sfrs << '\n';

  for (i = 0; i < num_of_sfrs; i++)
    {
      registers[sfr_map[i].address] = sfr_map[i].sfr_reg;
      registers[sfr_map[i].address]->address = sfr_map[i].address;
      registers[sfr_map[i].address]->value = sfr_map[i].por_value;
      registers[sfr_map[i].address]->break_point = 0;
      sfr_map[i].sfr_reg->cpu = this;
      // cout << "register: " << sfr_map[i].sfr_reg->name() << '\n';
    }

  option_reg.cpu = this;
  option_reg.value = 0xff;

  W.cpu = this;
}



//-------------------------------------------------------------------
//
// init_program_memory
//
// The purpose of this member function is to allocate memory for the
// pic's code space. The 'memory_size' parameter tells how much memory
// is to be allocated AND it should be an integer of the form of 2^n. 
// If the memory size is not of the form of 2^n, then this routine will
// round up to the next integer that is of the form 2^n.
//   Once the memory has been allocated, this routine will initialize
// it with the 'bad_instruction'. The bad_instruction is an instantiation
// of the instruction class that choke gpsim if it is executed. Note that
// there is only one instance of 'bad_instruction'.

void pic_processor::init_program_memory (unsigned int memory_size)
{

  pc.memory_size_mask = memory_size - 1;
  if ((pc.memory_size_mask) & memory_size)
    {
      cout << "*** WARNING *** memory_size should be of the form 2^N\n";

      memory_size = (memory_size + ~memory_size) & MAX_PROGRAM_MEMORY;
      pc.memory_size_mask = memory_size - 1;

      cout << "gpsim is rounding up to memory_size = " << memory_size << '\n';

    }

  // Initialize 'program_memory'. 'program_memory' is a pointer to an array of
  // pointers of type 'instruction'. This is where the simulated instructions
  // are stored.

  program_memory = (instruction **) new char[sizeof (instruction *) * memory_size];

  if (program_memory == NULL)
    {
      cout << "*** ERROR *** I can't get enough memory for the PIC program space\n";
      exit (1);
    }


  for (int i = 0; i < memory_size; i++)
    {
      program_memory[i] = &bad_instruction;
      program_memory[i]->cpu = this;     // %%% FIX ME %%% 
    }
}

//-------------------------------------------------------------------
//
void pic_processor::init_register_memory (unsigned int memory_size)
{

  // Allocate enough memory for the entire register space (e.g. 256 registers for 14-bit core)

  registers = (file_register **) new char[sizeof (file_register *) * FILE_REGISTERS];

  if (registers  == NULL)
    {
      cout << "*** ERROR *** I can't get enough memory for the PIC register space\n";
      exit (1);
    }

  // The register_bank points to the currently active bank of registers as defined
  // by the paging bits in the status register. Most of the pics have only 2 banks,
  // however some of the 5x parts have 4 and there's a possibility that this could
  // increase to 8 (if it hasn't already).

  register_bank = registers;

  // Make all of the file registers 'undefined' (each processor derived from this base
  // class defines its own register mapping).

  for (int i = 0; i < FILE_REGISTERS; i++)
    registers[i] = NULL;


}

//-------------------------------------------------------------------
//
//
// create_iopins
//
// The purpose of this member function is to create a set of nodes and to
// map them to the iopins on the processor. If stimuli are created during
// the simulation, then they are mapped to the iopin nodes that are created
// here.

void pic_processor::create_iopins (const IOPIN_map iopin_map[], unsigned int num_of_iopins)
{

  iopins = (IOPIN **) new char[sizeof (IOPIN *) * num_of_iopins];

  cout <<"creat iopins " << num_of_iopins << '\n';

  if (iopins  == NULL)
    {
      cout << "*** ERROR *** I can't get enough memory for IO nodes\n";
      exit (1);
    }


  int i;
  

  for(i=0; i<num_of_iopins; i++)
    {
      switch(iopin_map[i].type)
	{
	case BI_DIRECTIONAL_PU:
	  iopins[i] = new IO_bi_directional_pu(iopin_map[i].iop, iopin_map[i].iob);
	  break;

	case BI_DIRECTIONAL:
	  iopins[i] = new IO_bi_directional(iopin_map[i].iop, iopin_map[i].iob);
	  break;

	case INPUT_ONLY:
	  iopins[i] = new IO_input(iopin_map[i].iop, iopin_map[i].iob);
	  break;

	case OPEN_COLLECTOR:
	  iopins[i] = new IO_open_collector(iopin_map[i].iop, iopin_map[i].iob);
	  break;

	default:
	  cout << "Error: Invalid pin type in create_iopins.\n";
	}

    }

}


//-------------------------------------------------------------------
void pic_processor::dump_registers (void)
{
  unsigned int i, j, reg_num;
  unsigned int v;

  cout << "     " << hex;

  // Column labels
  for (i = 0; i < 16; i++)
    cout << setw(2) << setfill('0') <<  i << ' ';

  cout << '\n';

  for (i = 0; i < 16; i++)
    {
      cout << setw(2) << setfill('0') <<  i << ":  ";

      for (j = 0; j < 16; j++)
	{
	  reg_num = i * 16 + j;
	  if(registers[reg_num]->address == BAD_REGISTER)
	    cout << "-- ";
	  else
	    {
	      v = registers[reg_num]->get_value();
	      cout << setw(2) << setfill('0') <<  v << ' ';
	    }
	}
      cout << "   ";

      for (j = 0; j < 16; j++)
	{
	  reg_num = i * 16 + j;
	  v = registers[reg_num]->get_value();
	  if( (v >= ' ') && (v <= 'z'))
	    cout.put(v);
	  else
	    cout.put('.');
	}

      cout << '\n';

    }

  int regs_in_a_row = 0;

  int ad_hoc; // shit code to resolve register aliasing

  if(base_isa() == _12BIT_PROCESSOR_)
    ad_hoc = 0x20;
  else
    ad_hoc = 0x80;

  for (i = 0; i < 256; i++)
    {
      if(registers[i]->isa() == file_register::SFR_REGISTER)
	{
	  if((registers[i]->address != registers[i ^ ad_hoc]->address) ||
	     ( i < ad_hoc))
	    {
	      if(++regs_in_a_row > 4)
		{
		  cout << '\n';
		  regs_in_a_row = 1;
		}
	      int mask = ad_hoc - 1;
	      if(registers[i]->address != registers[i ^ ad_hoc]->address) 
		mask |= ad_hoc;


	      cout << setw(2) << (registers[i]->address & mask) << "  " <<  registers[i]->name() << " = " 
		   <<  setw(2) << setfill('0') <<  registers[i]->get_value();
	      int l = 10 - strlen(registers[i]->name());
	      for(int k=0; k<l; k++)
		cout << ' ';
	    }
	}
    }
  cout << "\nw = " << W.value << '\n';
  cout << "pc = " << pc.value << '\n';
  
  
}

//-------------------------------------------------------------------
void pic_processor::attach_src_line(int address,int file_id,int sline,int lst_line)
{

  if(address < program_memory_size())
    {
      program_memory[address]->update_line_number(file_id,sline,lst_line);

      if(sline > files[file_id].max_line)
	files[file_id].max_line = sline;

      if(lst_line+5 > files[lst_file_id].max_line)
	files[lst_file_id].max_line = lst_line+5;

    }

}

//-------------------------------------------------------------------
// read_src_files - this routine will open all of the source files 
//   associated with the project and associate their line numbers
//   with the addresses of the opcodes they generated.
//

void pic_processor::read_src_files(void)
{
  // Are there any src files ?
  for(int i=0; i<number_of_source_files; i++)
    {

      // did this src file generate any code?
      if(files[i].max_line > 0)
	{
	  // Create an array whose index corresponds to the
	  // line number of a source file line and whose data
	  // is the offset (in bytes) from the beginning of the 
	  // file. (e.g. files[3].line_seek[20] references the
	  // 20th line of the third source file.)
	  files[i].line_seek = new int[files[i].max_line];
	  if( NULL == (files[i].file_ptr = fopen(files[i].name,"r")))
	    continue;
	  rewind(files[i].file_ptr);

	  char buf[256],*s;
	  files[i].line_seek[0] = 0;
	  for(int j=1; j<=files[i].max_line; j++)
	    {
	      files[i].line_seek[j] = ftell(files[i].file_ptr);
	      s = fgets(buf,256,files[i].file_ptr);
	      if(s != buf)
		break;
	    }
	}
    }

}

//-------------------------------------------------------------------
void pic_processor::build_program_memory(int *memory,int minaddr, int maxaddr)
{
  if(maxaddr > program_memory_size()) {
    cout << "pic_processor::build_program_memory - maxaddr is tooooo big\n";
    cout << hex << maxaddr << "  " << program_memory_size() << '\n';
    return;
  }
  for (int i = minaddr; i < maxaddr; i++)
    {
      if(memory[i] != 0xffffffff)
	{
	  program_memory[i] = disasm(memory[i]);
	  if(program_memory[i] == NULL)
	    program_memory[i] = &bad_instruction;
	  program_memory[i]->add_line_number_symbol(i);
	}
    }
}

//-------------------------------------------------------------------
//
// load_hex
//
extern void readihex16 (FILE * file, int *memory, int *minaddr, int *maxaddr);

void pic_processor::load_hex (char *hex_file)
{


  ::config_word = 0xffff;   // default value

  FILE *inputfile = fopen (hex_file, "r");

  cout << "load hex\n";

  if (inputfile == NULL)
    {
      cout << "Error: Couldn't open " << hex_file << '\n';
      return;
    }

  int memory[MAXPICSIZE];
  int minaddr, maxaddr;

  // Fill the memory array with bogus data
  for(int i = 0; i<MAXPICSIZE; i++)
    memory[i] = 0xffffffff;

  readihex16 (inputfile, memory, &minaddr, &maxaddr);
  fclose (inputfile);

  build_program_memory(memory, minaddr, maxaddr);

  disassemble(minaddr, maxaddr);
  cout << "Configuration word = 0x"  << setw(4) << setfill('0') << ::config_word << '\n';

  set_frequency(10e6);
  set_config_word(::config_word);
  reset(POR_RESET);

  simulation_mode = STOPPED;

  cycles.dump_breakpoints();

}

