/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <iostream.h>
#include <iomanip.h>

#include "14bit-processors.h"
#include "trace.h"


Trace trace;     // Instantiate the trace buffer class


Trace::Trace(void)
{

  for(trace_index = 0; trace_index < TRACE_BUFFER_SIZE; trace_index++)
    trace_buffer[trace_index] = NOTHING;

  trace_index = 0;

}

void Trace::dump1(unsigned index)
{
  char a_string[50];
  unsigned int i;
  file_register *r;

  if(trace_buffer[index] & 0x80000000)
    {
      cout << "cycle: " << hex << (trace_buffer[index]&0x7fffffff) << '\n';
    }
  else
    {
      switch (trace_buffer[index] & 0xff000000)
	{
	case NOTHING:
	  cout << "empty trace cycle\n";
	  break;
	case INSTRUCTION:
	  cout << "instruction:" << hex << (trace_buffer[index]&0xffff) << '\n';
	  break;
	case PROGRAM_COUNTER:
	  i = trace_buffer[index]&0xffff;
	  cout << "pc:" << hex << setfill('0') << setw(4) << i 
	       << "  " << cpu->program_memory[i]->name(a_string) << '\n';
	  break;
	case PC_SKIP:
	  i = trace_buffer[index]&0xffff;
	  cout << "skipped:" << hex << setfill('0') << setw(4) << i 
	       << "  " << cpu->program_memory[i]->name(a_string) << '\n';
	  break;
	case REGISTER_READ:
	  r = cpu->registers[(trace_buffer[index]>>8) & 0xff];
	  cout << "read: " << hex << (trace_buffer[index]&0xff) << " from " << r->name() << '\n';
	  break;
	case REGISTER_WRITE:
	  r = cpu->registers[(trace_buffer[index]>>8) & 0xff];
	  cout << "wrote: " << hex << (trace_buffer[index]&0xff) << " to " << r->name() << '\n';
	  break;
	case READ_W:
	  cout << "read: " << hex << (trace_buffer[index]&0xff) << " from W\n";
	  break;
	case WRITE_W:
	  cout << "wrote: " << hex << (trace_buffer[index]&0xff) << " to W\n";
	  break;
	case WRITE_TRIS:
	  cout << "wrote: " << hex << (trace_buffer[index]&0xff) << " to TRIS\n";
	  break;
	case WRITE_OPTION:
	  cout << "wrote: " << hex << (trace_buffer[index]&0xff) << " to OPTION\n";
	  break;
	  //    case CYCLE_COUNTER:
	  //      break;
	case BREAKPOINT:
	  cout << "break: ";
	  bp.dump_traced(trace_buffer[index] & 0xffffff);
	  break;
	case INTERRUPT:
	  cout << "interrupt:\n";
	  break;
	case RESET:
	  switch( (RESET_TYPE) (trace_buffer[i]&0xff))
	    {
	    case POR_RESET:
	      cout << "POR";
	      break;
	    case WDT_RESET:
	      cout << "WDT";
	      break;
	    default:
	      cout << "unknown";
	    }
	  cout << " reset";
	  break;

	default:
	  cout << "*** INVALID TRACE ***\n";
	}
    }
}


void Trace::dump(unsigned int n=0)
{
  int i;

  if(n)
    i = (trace_index - n) & TRACE_BUFFER_MASK;
  else
    i = (trace_index - 10) & TRACE_BUFFER_MASK;

  do
    {
      dump1(i);
      i = (i+1) & TRACE_BUFFER_MASK;
    } while(i != trace_index);
}



void Trace::dump_last_instruction(void)
{
  char a_string[50];
  unsigned int i;
  unsigned int
    instruction_index,
    pc_index;
  bool 
    found_instruction = 0,
    found_pc =0;
  file_register *r;


  i = (trace_index-1) & TRACE_BUFFER_MASK;

  // Starting at the latest trace, search backwards first for an instruction and then for a program_counter
  do
  {
    switch (trace_buffer[i] & 0xff000000)
    {

    case INSTRUCTION:
      instruction_index = i;
      found_instruction = 1;
      break;

    case PROGRAM_COUNTER:
      if(found_instruction) 
	{
	  pc_index = i;
	  found_pc = 1;
	}
      break;

    }

    i = (i-1) & TRACE_BUFFER_MASK;

 
  // i == trace_index then we looped all the way around without finding anything
  // that should NEVER happen.
  
  } while( (i != trace_index) && (!found_pc));


  if(found_pc)
    {
      cout << "processor: " << cpu->name_str << "   ";
      //      cout << "i "<<i << "  trace index " << trace_index << "  inst "<< instruction_index << "  pc "<<pc_index << '\n';
      i = trace_buffer[pc_index]&0xffff;

      // Display the pc, opcode, and mnemonic
      cout << hex << 
	setfill('0') << setw(4) << i  << "  " << 
	setfill('0') << setw(4) << (trace_buffer[instruction_index]&0xffff) << "   "  << 
	(cpu->program_memory[i]->name(a_string)) << '\n';

      if(instruction_index != trace_index) 
	{
	  // Get the trace buffer index just past the point where the instruction was traced
	  i = (instruction_index  + 1) & TRACE_BUFFER_MASK;

	  // Now search the trace buffer for all register accesses
	  do
	    {
	      if(trace_buffer[i] & 0x80000000)
		{
		  cout << "cycle: " << hex << (trace_buffer[i]&0x7fffffff) << '\n';
		}
	      else
		{
		  switch (trace_buffer[i] & 0xff000000)
		    {
		    case REGISTER_READ:
		      r = cpu->registers[(trace_buffer[i]>>8) & 0xff];
		      cout << "read: " << hex << (trace_buffer[i]&0xff) << " from " << r->name() << '\n';
		      break;
		    case REGISTER_WRITE:
		      r = cpu->registers[(trace_buffer[i]>>8) & 0xff];
		      cout << "wrote: " << hex << (trace_buffer[i]&0xff) << " to " << r->name() << '\n';
		      break;
		    case READ_W:
		      cout << "read W: " << hex << (trace_buffer[i]&0xff) << '\n';
		      break;
		    case WRITE_W:
		      cout << "wrote W: " << hex << (trace_buffer[i]&0xff) << '\n';
		      break;
		    case WRITE_TRIS:
		      cout << "wrote TRIS: " << hex << (trace_buffer[i]&0xff) << '\n';
		      break;
		    case WRITE_OPTION:
		      cout << "wrote OPTION: " << hex << (trace_buffer[i]&0xff) << '\n';
		      break;

		    }
		}
	      i = (i+1) & TRACE_BUFFER_MASK;
	       
	    } while(i != trace_index);
	} 
      else
	{
	  cout << "nothing past the instruction   trace_index " << trace_index << "  inst index " << instruction_index << '\n';
	}

    }
  else
    {
      cout << "Error while dumping the trace (couldn't find any instructions)\n";
    }
}
